use bus::Bus;
use crate::{
    app::{
        App,
        authentication::Authentication,
        user::User,
        connection::Connection,
    },
    protocol,
};

pub async fn thread_system(mut app:App, bus:Bus<protocol::QRPacket>)
{
    use protocol::*;
    use protocol::QRPacketData::*;
    use ring::rand::{SecureRandom, SystemRandom};

    let rng = SystemRandom::new();
    let argon_config = argon2::Config::default();

    while match bus.receive_wait() {
        Some(packet) => {
            let qr = &packet.data;

            let mut user_id = None;
            if let Some(conn) = app.connections.get(qr.id as usize) {
                if let Some(auth_id) = conn.auth {
                    if let Some(auth) = app.auths.get(&auth_id) {
                        user_id = Some(auth.user);
                    }
                }
            }

            match &qr.data {
                QConn(request) => {
                    let id = app.connections.add(Connection {
                        bus: request.bus_id,
                        auth: None,
                    });

                    println!("Connect: {}", id);

                    bus.send(packet.from, QRPacket::new(id as u32, RConn)).is_ok()
                }

                QDisconn => {
                    app.connections.remove(qr.id as usize).ok();

                    println!("Disconnect: {}", qr.id);
                    true
                }

                QRegister(request) => {
                    let mut response = PacketRegisterResponse::new();
                    response.status = STATUS_ERROR;

                    println!("Request: Register");

                    let mut is_valid = true;
                    if request.code != "abc".as_bytes() { response.status = STATUS_BAD_CODE; is_valid = false; }
                    if is_valid && request.handle.len() == 0 { response.status = STATUS_BAD_HANDLE; is_valid = false; }
                    if is_valid && request.secret.len() == 0 { response.status = STATUS_BAD_SECRET; is_valid = false; }

                    if is_valid {
                        match app.user_handle.get(request.handle.as_bytes()) {
                            None => {
                                let mut salt = [0u8; 16];
                                match rng.fill(&mut salt) {
                                    Ok(_) => {
                                        let salt_id = app.salts.len();
                                        app.salts.push(salt);

                                        if let Ok(hash) = argon2::hash_raw(&request.secret, &salt, &argon_config) {
                                            let user_id = app.user_next;
                                            app.user_next += 1;

                                            // Create user entry
                                            let user_pos = app.users.add(User {
                                                id:user_id,
                                                handle:request.handle.clone(),
                                                secret:hash,
                                                na_key:salt_id,
                                            });

                                            // Register user pool id and handle
                                            app.user_id.set(user_id as isize, user_pos);
                                            app.user_handle.set(request.handle.as_bytes(), user_id);

                                            println!("Registered user '{}' @ {} with id {}", request.handle, user_pos, user_id);

                                            // Generate authentication token and secret
                                            response.status = STATUS_OK;
                                            rng.fill(&mut response.secret).ok();
                                            loop {
                                                rng.fill(&mut response.token).ok();

                                                if app.auths.get(&response.token).is_none() {
                                                    app.auths.set(&response.token, Authentication {
                                                        key:response.token,
                                                        secret:response.secret,
                                                        user:user_id,
                                                    });
                                                    break;
                                                }
                                            }
                                            
                                            // Attach authentication to connection
                                            if let Some(conn) = app.connections.get_mut(qr.id as usize) {
                                                conn.auth = Some(response.token);
                                            }
                                        }
                                    }
                                    Err(_) => { println!("error: failed to generate salt.") }
                                }
                            }
                            Some(_) => {
                                response.status = STATUS_BAD_HANDLE;
                                println!("notice: attempt to register existing handle: '{}'", request.handle);
                            }
                        }
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RRegister(response))).is_ok()
                }

                QAuth(request) => {
                    let mut response = PacketAuthResponse::new();
                    response.status = STATUS_ERROR;

                    println!("Request: Auth");

                    let mut is_valid = true;
                    if is_valid && request.handle.len() == 0 { response.status = STATUS_BAD_HANDLE; is_valid = false; }
                    if is_valid && request.secret.len() == 0 { response.status = STATUS_BAD_SECRET; is_valid = false; }

                    if is_valid {

                        // Get user data from handle
                        match app.user_handle.get(request.handle.as_bytes()) {
                            Some(uid) => {
                                if let Some(tuid) = app.user_id.get(*uid as isize) {
                                    if let Some(user) = app.users.get(*tuid) {
                                        // Get user salt
                                        if let Some(salt) = app.salts.get(user.na_key) {

                                            // Verify salted secret against user data
                                            if argon2::verify_raw(&request.secret.as_bytes(), salt, &user.secret, &argon_config).unwrap_or(false) {
                                                println!("Authenticated user '{}' id {}", request.handle, *uid);

                                                // Generate authentication token and secret
                                                response.status = STATUS_OK;
                                                rng.fill(&mut response.secret).ok();
                                                loop {
                                                    rng.fill(&mut response.token).ok();

                                                    if app.auths.get(&response.token).is_none() {
                                                        app.auths.set(&response.token, Authentication {
                                                            key:response.token,
                                                            secret:response.secret,
                                                            user:*uid,
                                                        });
                                                        break;
                                                    }
                                                }
                                                
                                                // Attach authentication to connection
                                                if let Some(conn) = app.connections.get_mut(qr.id as usize) {
                                                    conn.auth = Some(response.token);
                                                }
                                            } else {
                                                println!("notice: password verification failed.");
                                            }
                                        } else {
                                            println!("error: user salt id '{}' not found.", user.na_key);
                                        }
                                    } else {
                                        println!("error: user with id '{}' not found.", uid);
                                    }
                                } else {
                                    println!("error: user with id '{}' not found.", uid);
                                }
                            }
                            None => { }
                        }
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RAuth(response))).is_ok()
                }

                QAuthResume(request) => {
                    let mut response = PacketAuthResumeResponse::new();
                    response.status = STATUS_ERROR;

                    if let Some(auth) = app.auths.get(&request.token) {

                        // Compare full secret length to reduce time-based attacks.
                        let mut valid = true;
                        for i in 0..16 {
                            valid |= auth.secret[i] == request.secret[i];
                        }

                        if valid {
                            if let Some(conn) = app.connections.get_mut(qr.id as usize) {
                                conn.auth = Some(request.token);
                                response.status = STATUS_OK;
                            } else {
                                response.status = STATUS_SERVER_ERROR;
                            }
                        }
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RAuthResume(response))).is_ok()
                }

                QDeauth => {
                    if let Some(conn) = app.connections.get_mut(qr.id as usize) {
                        match conn.auth {
                            Some(auth) => {
                                println!("Deauthenticated connection: {}", qr.id);
                                app.auths.unset(&auth);
                            }
                            None => { }
                        }
                        conn.auth = None;
                    }
                    true
                }

                QSessionList(request) => {
                    use game::game::GameState;
                    
                    println!("Request: Session List");

                    let mut response = PacketSessionListResponse::new();

                    let mut count = 0;

                    let mut next_id = app.session_time.begin();
                    while let Some(id) = next_id {
                        let token = app.session_time.get(id).unwrap();
                        if let Some(session) = app.sessions.get(token) {

                            // Requirements:
                            //  - GameState must be None or match state of session.
                            //  - Joinable must have either player slot empty.
                            //  - IsPlayer must have the current user in either player slot.
                            //  - IsLive must have both users connected to the session.

                            let mut valid = request.game_state == GameState::None || request.game_state == session.game.state;
                            valid &= request.game_state != GameState::Joinable || session.p_dawn.user.is_none() || session.p_dusk.user.is_none();
                            valid &= !request.is_player || session.p_dawn.user == user_id || session.p_dusk.user == user_id;
                            valid &= !request.is_live || (session.p_dawn.connection.is_some() && session.p_dusk.connection.is_some());

                            if valid {
                                let is_player = user_id.is_some() && (session.p_dawn.user == user_id || session.p_dusk.user == user_id);

                                let dawn_handle = if let Some(uid) = session.p_dawn.user {
                                    if let Some(cuid) = app.user_id.get(uid as isize) {
                                        if let Some(user) = app.users.get(*cuid) {
                                            user.handle.clone()
                                        } else { String::new() }
                                    } else { String::new() }
                                } else { String::new() };

                                let dusk_handle = if let Some(uid) = session.p_dusk.user {
                                    if let Some(cuid) = app.user_id.get(uid as isize) {
                                        if let Some(user) = app.users.get(*cuid) {
                                            user.handle.clone()
                                        } else { String::new() }
                                    } else { String::new() }
                                } else { String::new() };

                                response.records.push(PacketSessionListResponseRecord {
                                    token:session.key,
                                    handles:[
                                        dawn_handle,
                                        dusk_handle,
                                    ],
                                    turn:0,
                                    last_move:[0; 5],
                                    viewers:0,
                                    player:is_player,
                                });

                                count += 1;
                            }
                        }

                        if count >= 60 { break; }
                        next_id = app.session_time.next(id);
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RSessionList(response))).is_ok()
                }

                QSessionCreate(_request) => {
                    use crate::app::session::*;

                    println!("Request: Session Create");

                    let mut response = PacketSessionCreateResponse::new();
                    response.status = STATUS_ERROR;

                    if let Some(uid) = user_id {
                        // Generate session token
                        let mut token = SessionToken::default();
                        let mut secret = SessionSecret::default();
                        loop {
                            rng.fill(&mut token).ok();
                            if app.sessions.get(&token).is_none() { break; }
                        }
                        rng.fill(&mut secret).ok();

                        let chain_id = app.session_time.add(token);
                        app.sessions.set(&token, Session {
                            key:token,
                            secret:secret,
                            game:game::Game::new(),
                            p_dawn:Viewer {
                                connection:None,//Some(qr.id),
                                user:Some(uid),
                            },
                            p_dusk:Viewer {
                                connection:None,
                                user:None,
                            },
                            viewers:Vec::new(),
                            chain_id:chain_id,
                        });
                        app.session_time.set(chain_id, token);

                        // Set player to Dawn.
                        response.mode = 0;

                        response.status = STATUS_OK;
                        response.token = token;
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RSessionCreate(response))).is_ok()
                }

                QSessionJoin(request) => {
                    println!("Request: Session Join");

                    let mut response = PacketSessionJoinResponse::new();
                    response.status = STATUS_ERROR;

                    // Verify that session exists.
                    if let Some(session) = app.sessions.get_mut(&request.token) {

                        // Join game as player.
                        if request.join {

                            // Verify client is authenticated.
                            if let Some(uid) = user_id {
                                
                                // User must not already be player.
                                if session.p_dawn.user != user_id && session.p_dusk.user != user_id {

                                    // Add user to empty player slot.
                                    if if session.p_dawn.user.is_none() {
                                        session.p_dawn.user = Some(uid);
                                        response.mode = 0;
                                        true
                                    } else if session.p_dusk.user.is_none() {
                                        session.p_dusk.user = Some(uid);
                                        response.mode = 1;
                                        true
                                    } else {
                                        // Session is not empty.
                                        response.status = STATUS_ERROR;
                                        false
                                    } {
                                        println!("Add user to session.");
                                        response.status = STATUS_OK;
                                    }
                                } else {
                                    println!("User resumes session.");
                                    response.status = STATUS_OK;
                                    response.mode = (session.p_dusk.user == user_id) as u8;
                                }
                            } else { response.status = STATUS_NOAUTH; }
                        }

                        // Join game as spectator.    
                        else {
                            println!("User spectates session.");
                            response.status = STATUS_OK;
                            response.mode = 2;
                        }
                    }

                    bus.send(packet.from, QRPacket::new(qr.id, RSessionJoin(response))).is_ok()
                }

                _ => { true }
            }
        }
        None => false,
    } { }
}
