use std::{io::Read, sync::{Arc, RwLock}};

use crate::util::string::minimize_whitespace;

struct WebCacheData {
    html:String,
    css:String,
    js:String,
    favicon:Vec<u8>,
}

#[derive(Clone)]
pub struct WebCache {
    data:Arc<RwLock<WebCacheData>>,
}
impl WebCache {
    pub fn init() -> Self
    {
        use std::fs::File;

        let mut html = String::new();
        let mut favicon = Vec::<u8>::new();

        // Cache html file
        match File::open("www/.html") {
            Ok(mut file) => {
                file.read_to_string(&mut html).ok();
                html = minimize_whitespace(&html);
            }
            Err(_) => { }
        }

        // Cache js file
        let css_path = std::path::Path::new("www/css/");
        let css: String = [
            "main.css",
            "util.css",
            "ui.css",
            "form.css",
            "game.css",
        ].map(|path| {
            let mut buffer = String::new();
            match File::open(css_path.join(path)) {
                Ok(mut file) => { file.read_to_string(&mut buffer).ok(); }
                Err(_) => { }
            }
            buffer
        }).concat();

        // Cache js file
        let js_path = std::path::Path::new("www/js/");
        let js = [
            "const.js",
            "util.js",
            "game_asset.js",
            "game.js",
            "interface.js",
            "ui.js",
            "scene.js",
            "system.js",
            "main.js",
        ].map(|path| {
            let mut buffer = String::new();
            match File::open(js_path.join(path)) {
                Ok(mut file) => { file.read_to_string(&mut buffer).ok(); }
                Err(_) => { }
            }
            buffer
        }).concat();

        // Cache favicon file
        match File::open("www/favicon.png") {
            Ok(mut file) => {
                file.read_to_end(&mut favicon).ok();
            }
            Err(_) => { }
        }

        Self {
            data:Arc::new(RwLock::new(WebCacheData {
                html, css, js, favicon,
            })),
        }
    }

    pub fn html(&self) -> String
    {
        match self.data.read() {
            Ok(reader) => reader.html.to_string(),
            Err(_) => String::new(),
        }
    }

    pub fn css(&self) -> String
    {
        match self.data.read() {
            Ok(reader) => reader.css.to_string(),
            Err(_) => String::new(),
        }
    }

    pub fn js(&self) -> String
    {
        match self.data.read() {
            Ok(reader) => reader.js.to_string(),
            Err(_) => String::new(),
        }
    }

    pub fn favicon(&self) -> Vec<u8>
    {
        match self.data.read() {
            Ok(reader) => reader.favicon.clone(),
            Err(_) => Vec::new(),
        }
    }
}
